from datetime import datetime,timedelta
import dateutil.parser
import dateutil.rrule
import numpy as np
import pandas as pd
import xarray as xr
import matplotlib as mpl
import matplotlib.pyplot as plt
import types
import xml.etree.ElementTree as ET
from joblib import Parallel, delayed
from . import base,Utils
originData=base.originData
isInt=Utils.isInt
parse_element=Utils.parse_element

#region 绘图参数
ref_colors=[
    '#FFFFFF',
    '#000080',
    '#042AC9',
    '#0852D1',
    '#0C7AD5',
    '#01A0F6',
    '#00ECEC',
    '#00D800',
    '#019000',
    '#FFFF00',
    '#E7C000',
    '#FF9000',
    '#FF0000',
    '#D60000',
    '#FFFFFF'
]
ref_levels=[-30,-20,-10,-5,0,5,10,15,20,25,30,35,40]
ref_cmap = (mpl.colors.ListedColormap(ref_colors[1:-1]).with_extremes(over=ref_colors[-1], under=ref_colors[0]))
ref_norm = mpl.colors.BoundaryNorm(ref_levels, ref_cmap.N)

velocity_colors=[
    '#FFFFFF',
    '#042948',
    '#053761',
    '#064275',
    '#094C85',
    '#0A5797',
    '#0B62AB',
    '#0B6EBF',
    '#0C7AD5',
    '#01A0F6',
    '#80E3FF',
    '#00E0FE',
    '#00B0B0',
    '#00FE00',
    '#00C400',
    '#008000',
    '#D8D2D8',
    '#E7E3E8',
    '#FE0000',
    '#FE5858',
    '#FEB0B0',
    '#FE7C00',
    '#FED200',
    '#FEFE00',
    '#9B0CE2',
    '#8B0CCB',
    '#7E0BB7',
    '#6F0AA2',
    '#620A8F',
    '#57087F',
    '#4A076C',
    '#3D0758',
    '#2E0543',
    "#FFFFFF",
]
velocity_levels=[-18,-16,-15,-14,-13,-12,-11,-9,-8,-7,-6,-5,-4,-3,-2,-1,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15]
velocity_cmap = (mpl.colors.ListedColormap(velocity_colors[1:-1]).with_extremes(over=velocity_colors[-1], under=velocity_colors[0]))
velocity_norm = mpl.colors.BoundaryNorm(velocity_levels, velocity_cmap.N)

spectrumwith_colors=[
    "#FFFFFF",
    "#E7E3E7",
    "#7BE3E7",
    "#00E3E7",
    "#00B2B5",
    "#00FFFF",
    "#00C700",
    "#008200",
    "#FFFF00",
    "#FFD300",
    "#FF7D00",
    "#FFB2B5",
    "#AD595A",
    "#75005F",
    "#FFFFFF",
]
spectrumwith_levels=[0,0.5,1,1.5,2,2.5,3,3.5,4,4.5,5,5.5,6]
spectrumwith_cmap = (mpl.colors.ListedColormap(spectrumwith_colors[1:-1]).with_extremes(over=spectrumwith_colors[-1], under=spectrumwith_colors[0]))
spectrumwith_norm = mpl.colors.BoundaryNorm(spectrumwith_levels, spectrumwith_cmap.N)

snr_colors=[
    "#FFFFFF",
    "#0535CD",
    "#0A67D3",
    "#109AD8",
    "#15CCDE",
    "#1AFFE3",
    "#48FFB6",
    "#76FF88",
    "#A3FF5B",
    "#D1FF2D",
    "#FFFF00",
    "#F5CC00",
    "#EC9900",
    "#E26601",
    "#D93301",
    "#FFFFFF",
]
snr_levels=[-30,-25,-20,-15,-10,-5,0,5,10,15,20,25,30,35,40]
snr_cmap = (mpl.colors.ListedColormap(snr_colors[1:-1]).with_extremes(over=snr_colors[-1], under=snr_colors[0]))
snr_norm = mpl.colors.BoundaryNorm(snr_levels, snr_cmap.N)

#endregion

#region 元数据配置
GenericHeader=np.dtype(
     [
         ('MagicNumber', 'i4'),
         ('MajorVersion', 'i2'),
         ('MinorVersion', 'i2'),
         ('GenericType', 'i4'),
         ('Reserved', 'S20'),
     
     ]
)
SiteConfig=np.dtype(
     [
         ('Site_Code', 'S8'),
         ('Site_Name', 'S24'),
         ('Latitude', 'f4'),
         ('Longitude', 'f4'),
         ('Antenna_Height', 'f4'),
         ('Ground_Height', 'f4'),
         ('Amend_North', 'f4'),
         ('RDA_Version', 'i2'),
         ('Radar_Type', 'i2'),
         ('Manufacturers', 'S6'),
         ('Reserved', 'S10'),
     
     ]
)
RadarConfig=np.dtype(
     [
         ('Frequency', 'f4'),
         ('Wavelength', 'f4'),
         ('Beam_Width_Hori', 'f4'),
         ('Beam_Width_Vert', 'f4'),
         ('Transmitter_peak_power', 'f4'),
         ('Antenna_gain', 'f4'),
         ('Total_loss', 'f4'),
         ('Receiver_gain', 'f4'),
         ('First_side', 'f4'),
         ('Receiver_dynamic_Range', 'f4'),
         ('Receiver_Sensitivity', 'f4'),
         ('Band_Width', 'f4'),
         ('Max_Explore_Range', 'u4'),
         ('Distance_solution', 'u2'),
         ('Polarization_Type', 'u2'),
         ('Reserved', 'S96')
     
     ]
)
TaskConfig=np.dtype(
     [
         ('Task_Name', 'S16'),
         ('Task_Description', 'S96'),
         ('Polarization_Way', 'i2'),
         ('Scan_Type', 'i2'),
         ('Pulse_Width_1', 'i4'),
         ('Pulse_Width_2', 'i4'),
         ('Pulse_Width_3', 'i4'),
         ('Pulse_Width_4', 'i4'),
         ('Scan_Start_Time', 'u8'),
         ('Cut_Number', 'i4'),
         ('Horizontal_Noise', 'f4'),
         ('Vertical_Noise', 'f4'),
         ('Horizontal_Calibration1', 'f4'),
         ('Horizontal_Calibration2', 'f4'),
         ('Horizontal_Calibration3', 'f4'),
         ('Horizontal_Calibration4', 'f4'),
         ('Vertical_Calibration1', 'f4'),
         ('Vertical_Calibration2', 'f4'),
         ('Vertical_Calibration3', 'f4'),
         ('Vertical_Calibration4', 'f4'),
         ('Horizontal_Noise_Temperature', 'f4'),
         ('Vertical_Noise_Temperature', 'f4'),
         ('ZDR_Calibration', 'f4'),
         ('PHIDP_Calibration', 'f4'),
         ('LDR_Calibration', 'f4'),
         ('Number_of_coherent_accumulation_1', 'S1'),
         ('Number_of_coherent_accumulation_2', 'S1'),
         ('Number_of_coherent_accumulation_3', 'S1'),
         ('Number_of_coherent_accumulation_4', 'S1'),
         ('FFT_Count_1', 'u2'),
         ('FFT_Count_2', 'u2'),
         ('FFT_Count_3', 'u2'),
         ('FFT_Count_4', 'u2'),
         ('Accumulation_of_power_spectrum_1', 'S1'),
         ('Accumulation_of_power_spectrum_2', 'S1'),
         ('Accumulation_of_power_spectrum_3', 'S1'),
         ('Accumulation_of_power_spectrum_4', 'S1'),
         ('Pulse_width_1_starting_position', 'u4'),
         ('Pulse_width_2_starting_position', 'u4'),
         ('Pulse_width_3_starting_position', 'u4'),
         ('Pulse_width_4_starting_position', 'u4'),
         ('Reserved', 'S20'),
     ]
)
CutConfig=np.dtype([
    ('Process_Mode', 'i2'),  # SHORT
    ('Wave_Form', 'i2'),  # SHORT
    ('PRF_1', 'f4'),  # FLOAT
    ('PRF_2', 'f4'),  # FLOAT
    ('PRF_3', 'f4'),  # FLOAT
    ('PRF_4', 'f4'),  # FLOAT
    ('PRF_Mode', 'i2'),  # SHORT
    ('Pulse_width_combination_mode', 'i2'),  # SHORT
    ('Azimuth', 'f4'),  # FLOAT
    ('Elevation', 'f4'),  # FLOAT
    ('Start_Angle', 'f4'),  # FLOAT
    ('End_Angle', 'f4'),  # FLOAT
    ('Angular_Resolution', 'f4'),  # FLOAT
    ('Scan_Speed', 'f4'),  # FLOAT
    ('Log_Resolution', 'i4'),  # INT
    ('Doppler_Resolution', 'i4'),  # INT
    ('Start_Range', 'i4'),  # INT
    ('Phase_Mode', 'i4'),  # INT
    ('Atmospheric_Loss', 'f4'),  # FLOAT
    ('Nyquist_Speed', 'f4'),  # FLOAT
    ('Misc_Filter_Mask', 'i4'),  # INT
    ('SQI_Threshold', 'f4'),  # FLOAT
    ('SIG_Threshold', 'f4'),  # FLOAT
    ('CSR_Threshold', 'f4'),  # FLOAT
    ('LOG_Threshold', 'f4'),  # FLOAT
    ('CPA_Threshold', 'f4'),  # FLOAT
    ('PMI_Threshold', 'f4'),  # FLOAT
    ('DPLOG_Threshold', 'f4'),  # FLOAT
    ('Thresholds_r', 'S12'),  # CAHR*12 (12 Bytes)
    ('dBT_Mask', 'i4'),  # INT
    ('dBZ_Mask', 'i4'),  # INT
    ('Velocity_Mask', 'i4'),  # INT
    ('Spectrum_Width_Mask', 'i4'),  # INT
    ('DP_Mask', 'i4'),  # INT
    ('Mask_Reserved', 'S12'),  # 12 Bytes
    ('Scan_Sync', 'i4'),  # INT
    ('Direction', 'i4'),  # INT
    ('Ground_Clutter_Classifier_Type', 'i2'),  # SHORT
    ('Ground_Clutter_Filter_Type', 'i2'),  # SHORT
    ('Ground_Clutter_Filter_Notch_Width', 'i2'),  # SHORT (0.1 m/s)
    ('Ground_Clutter_Filter_Window', 'i2'),  # SHORT
    ('Reserved', 'S92')  # 92 Bytes
])
RadialHeader=np.dtype([
    ('Radial_State', 'i2'),  # SHORT
    ('Spot_Blank', 'i2'),  # SHORT
    ('Sequence_Number', 'u2'),  # USHORT
    ('Radial_Number', 'u2'),  # USHORT
    ('Moment_Number', 'u2'),  # USHORT
    ('Elevation_Number', 'u2'),  # USHORT
    ('Azimuth', 'f4'),  # FLOAT
    ('Elevation', 'f4'),  # FLOAT
    ('Seconds', 'u8'),  # ULONG
    ('Microseconds', 'u4'),  # UINT
    ('Length_of_data', 'u4'),  # UINT
    ('Duration_Seconds', 'u2'),  # USHORT
    ('Max_FFT_Count', 'u2'),  # USHORT
    ('Reserved', 'S24')  # 24 Bytes
])
data_unit_header=np.dtype([
    ('Data_Type', 'u2'),  # USHORT
    ('Scale', 'u2'),  # USHORT
    ('Offset', 'u2'),  # USHORT
    ('Bin_Bytes', 'u2'),  # USHORT
    ('Bin_Number', 'u2'),  # USHORT
    ('Flags', 'i2'),  # SHORT
    ('Data_Length', 'i4'),  # INT
    ('Reserved', 'S16')  # 16 Bytes
])
FFT_extra_unit_header = [
    ['FFT_Count', 'u2', [1024,]],  # SHORT* L
    ['Number_of_coherent_accumulation', 'u1', [1024,]],  # CHAR* L
    ['Waveform_Number', 'u1', [1024,]],  # CHAR* L
    ['Accumulation_of_power_spectrum', 'u1', [1024,]]  # CHAR* L
]
#endregion

SingleFFTData=types.new_class('SingleFFTData',(originData,))
FFTDatas=types.new_class('FFTDatas',(originData,))
SingleBaseData=types.new_class('SingleBaseData',(originData,))
BaseDatas=types.new_class('BaseDatas',(originData,))
StatusData=types.new_class('StatusData',(originData,))
CalibrationData=types.new_class('CalibrationData',(originData,))

unobdata=999999
nodata=np.nan

def readSingleFFTData(fp:str)->SingleFFTData:
    '''
    读取云雷达FFT数据文件（秒），文件命名格式为：Z_RADA_I_IIiii_yyyyMMddhhmmss_O_YCCR_设备型号_FFT_MM.BIN
    Args:
        fp:文件路径

    Returns:
        singleFFTData:云雷达FFT数据对象
    '''
    FFTScale=100.0
    FFTOffset=32002.0
    with open(fp,'rb') as f:
        bs=f.read()
    yldfft=SingleFFTData()
    bsoffset_left=0
    bsoffset_right=bsoffset_left+GenericHeader.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], GenericHeader)
    yldfft['GenericHeader'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    bsoffset_left=bsoffset_right
    bsoffset_right=bsoffset_left+SiteConfig.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], SiteConfig)
    yldfft['SiteConfig']  = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    bsoffset_left=bsoffset_right
    bsoffset_right=bsoffset_left+RadarConfig.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], RadarConfig)
    yldfft['RadarConfig'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    bsoffset_left=bsoffset_right
    bsoffset_right=bsoffset_left+TaskConfig.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], TaskConfig)
    yldfft['TaskConfig'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    yldfft['TaskConfig']['Scan_Start_Time']=yldfft['TaskConfig']['Scan_Start_Time'].astype('datetime64[s]')
    Cut_Number=yldfft['TaskConfig']['Cut_Number']
    yldfft['CutConfigs']=[]
    for i in range(Cut_Number):
        bsoffset_left = bsoffset_right
        bsoffset_right = bsoffset_left + CutConfig.itemsize
        data=np.frombuffer(bs[bsoffset_left:bsoffset_right], CutConfig)
        yldfft['CutConfigs'].append(originData.from_dict({name: data[name][0] for name in data.dtype.names}))
    bsoffset_left = bsoffset_right
    bsoffset_right = bsoffset_left + RadialHeader.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], RadialHeader)
    yldfft['RadialHeader'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    yldfft['DataInfos']=[]
    yldfft['Data']=[]
    for i in range(yldfft['RadialHeader']['Moment_Number']):
        bsoffset_left = bsoffset_right
        bsoffset_right = bsoffset_left + data_unit_header.itemsize
        data=np.frombuffer(bs[bsoffset_left:bsoffset_right], data_unit_header)
        RadialDatai=originData.from_dict({name: data[name][0] for name in data.dtype.names})
        FFT_extra_unit_headeri=FFT_extra_unit_header.copy()
        for i in range(len(FFT_extra_unit_headeri)):
            FFT_extra_unit_headeri[i][2]=RadialDatai['Bin_Number']
            FFT_extra_unit_headeri[i]=tuple(FFT_extra_unit_headeri[i])
        FFT_extra_unit_headeri=np.dtype(FFT_extra_unit_headeri)
        bsoffset_left = bsoffset_right
        bsoffset_right = bsoffset_left + FFT_extra_unit_headeri.itemsize
        data=np.frombuffer(bs[bsoffset_left:bsoffset_right], FFT_extra_unit_headeri)
        RadialDatai.update({name: data[name][0] for name in data.dtype.names})
        yldfft['DataInfos'].append(RadialDatai)
        bsoffset_left = bsoffset_right
        bsoffset_right = bsoffset_left +int(RadialDatai['Bin_Bytes'])*int(RadialDatai['Bin_Number'])
        yldfft['Data'].append((np.frombuffer(
            bs[bsoffset_left:bsoffset_right],
            dtype='u2',count=2*int(yldfft.RadialHeader.Max_FFT_Count)*int(RadialDatai.Bin_Number)
        ).reshape((int(RadialDatai.Bin_Number),int(yldfft.RadialHeader.Max_FFT_Count),2))-FFTOffset)/FFTScale)
    data=np.array(yldfft['Data'])
    data=np.expand_dims(data,1)
    heights = np.arange(
        yldfft['CutConfigs'][0]['Start_Range'],
        yldfft['CutConfigs'][0]['Start_Range'] + yldfft['CutConfigs'][0]['Doppler_Resolution'] * yldfft.DataInfos[0][
            'Bin_Number'],
        yldfft['CutConfigs'][0]['Log_Resolution']
    )
    yldfft['Data']=xr.Dataset(
        data_vars={f'FFT{i}':(['time','height','FFT_index','dtype'],data[i]) for i in range(yldfft['RadialHeader']['Moment_Number'])},
        coords={
            'time':[yldfft['TaskConfig']['Scan_Start_Time'].astype(datetime)],
            'height': heights,
            'index':range(yldfft['RadialHeader']['Max_FFT_Count']),
            'dtype':['flag','value']
        },
        attrs={
            'time_count':1,
            'height_count': len(heights),
            'time_reference': 'UTC',
            'height_unit': 'meter',
        }
    )
    return yldfft

def FFTDatas_getDatas(self,fixData_Length='max',unobdata=unobdata)->xr.Dataset:
    '''
    获取云雷达基数据集中的数据
    Args:
        fixData_Length:不同时次数据对齐方式，max:最大长度,min:最小长度,或者指定长度
    Returns:
        xr.Dataset:数据
    '''
    varnames = list(map(lambda di: list(di.Data.data_vars), self.FFTDatas))
    varnames = list(set([item for sublist in varnames for item in sublist]))
    hcounts = list(map(lambda di: di.Data.height_count, self.FFTDatas))
    times = list(map(lambda di: di.Data.time.values[0], self.FFTDatas))
    if(fixData_Length=='max'):
        maxlendatas = max(hcounts)
    if(fixData_Length=='min'):
        maxlendatas = min(hcounts)
    if (isInt(fixData_Length)):
        maxlendatas = int(fixData_Length)
    heights = np.arange(
        self.FFTDatas[0]['CutConfigs'][0]['Start_Range'],
        self.FFTDatas[0]['CutConfigs'][0]['Start_Range'] + self.FFTDatas[0]['CutConfigs'][0][
            'Doppler_Resolution'] * maxlendatas,
        self.FFTDatas[0]['CutConfigs'][0]['Log_Resolution']
    )
    indexs=range(self.FFTDatas[0]['RadialHeader']['Max_FFT_Count'])
    datavars = {}
    for key in varnames:
        dds = list(map(lambda x: np.squeeze(x.Data[key].values).tolist(), filter(lambda x: key in x.Data, self.FFTDatas)))
        ddas = [
            list(d)[:maxlendatas] if (len(d) > maxlendatas) else list(d) + [unobdata] * (maxlendatas - len(d))
            for d in dds
        ]
        datavars[key] = (['time','height','FFT_index','dtype'], np.array(ddas),)
    Datas = xr.Dataset(
        data_vars=datavars,
        coords={
            'time': times,
            'height': heights,
            'index':indexs,
            'dtype':['flag','value']
        },
        attrs={
            'time_count':len(times),
            'height_count': maxlendatas,
            'time_reference': 'UTC',
            'height_unit': 'meter',
        }
    )
    self.Datas=Datas
FFTDatas.getDatas=FFTDatas_getDatas

def readFFTDatas(fps:list,use_multiprocess=False,multiproces_corenum=-1)->FFTDatas:
    '''
    读取云雷达谱数据文件列表
    Args:
        fps:文件路径列表
        use_multiprocess:是否使用多进程读取（速度较快但默认不使用）
        multiproces_corenum:多进程核心数（默认为-1，使用全部核心）
    Returns:
        list:云雷达对象列表
    '''
    rbds=FFTDatas()
    if(use_multiprocess):
        rbds['FFTDatas']=Parallel(n_jobs=multiproces_corenum)(delayed(readSingleFFTData)(fp) for fp in fps)
    else:
        rbds['FFTDatas']=[readSingleFFTData(fp) for fp in fps]
    rbds.__datas__=rbds['FFTDatas']

    rbds.getDatas()
    return rbds

def readSingleBaseData(fp:str)->SingleBaseData:
    '''    
    读取云雷达基数据文件（分钟），文件命名格式为：Z_RADA_I_IIiii_yyyyMMddhhmmss_O_YCCR_设备型号_RAW_MM.BIN
    Args:
        fp:文件路径
    Returns:
        SingleBaseData:云雷达基数据对象
    '''
    with open(fp,'rb') as f:
        bs=f.read()
    yldbd=SingleBaseData()
    bsoffset_left=0
    bsoffset_right=bsoffset_left+GenericHeader.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], GenericHeader)
    yldbd['GenericHeader'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    bsoffset_left=bsoffset_right
    bsoffset_right=bsoffset_left+SiteConfig.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], SiteConfig)
    yldbd['SiteConfig']  = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    bsoffset_left=bsoffset_right
    bsoffset_right=bsoffset_left+RadarConfig.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], RadarConfig)
    yldbd['RadarConfig'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    bsoffset_left=bsoffset_right
    bsoffset_right=bsoffset_left+TaskConfig.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], TaskConfig)
    yldbd['TaskConfig'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    yldbd['TaskConfig']['Scan_Start_Time']=yldbd['TaskConfig']['Scan_Start_Time'].astype('datetime64[s]')
    Cut_Number=yldbd['TaskConfig']['Cut_Number']
    yldbd['CutConfigs']=[]
    for i in range(Cut_Number):
        bsoffset_left = bsoffset_right
        bsoffset_right = bsoffset_left + CutConfig.itemsize
        data=np.frombuffer(bs[bsoffset_left:bsoffset_right], CutConfig)
        yldbd['CutConfigs'].append(originData.from_dict({name: data[name][0] for name in data.dtype.names}))
    bsoffset_left = bsoffset_right
    bsoffset_right = bsoffset_left + RadialHeader.itemsize
    data=np.frombuffer(bs[bsoffset_left:bsoffset_right], RadialHeader)
    yldbd['RadialHeader'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
    yldbd['Data']=[]
    yldbd['DataInfos']=[]
    for i in range(yldbd['RadialHeader']['Moment_Number']):
        bsoffset_left = bsoffset_right
        bsoffset_right = bsoffset_left + data_unit_header.itemsize
        data=np.frombuffer(bs[bsoffset_left:bsoffset_right], data_unit_header)
        datainfo=originData.from_dict({name: data[name][0] for name in data.dtype.names})
        data_type_value = datainfo['Data_Type']
        if data_type_value == 1:
            datainfo['Data_Name'] = 'Z1'
        elif data_type_value == 2:
            datainfo['Data_Name'] = 'V1'
        elif data_type_value == 3:
            datainfo['Data_Name'] = 'W1'
        elif data_type_value == 4:
            datainfo['Data_Name'] = 'SNR1'
        elif data_type_value == 5:
            datainfo['Data_Name'] = 'FFT1'
        elif data_type_value == 6:
            datainfo['Data_Name'] = 'Zc1'
        elif 6 <= data_type_value <= 16:
            datainfo['Data_Name'] = 'Reserved'
        elif data_type_value == 17:
            datainfo['Data_Name'] = 'Z2'
        elif data_type_value == 18:
            datainfo['Data_Name'] = 'V2'
        elif data_type_value == 19:
            datainfo['Data_Name'] = 'W2'
        elif data_type_value == 20:
            datainfo['Data_Name'] = 'SNR2'
        elif data_type_value == 21:
            datainfo['Data_Name'] = 'FFT2'
        elif data_type_value == 22:
            datainfo['Data_Name'] = 'Zc2'
        elif 22 <= data_type_value <= 32:
            datainfo['Data_Name'] = 'Reserved'
        elif data_type_value == 33:
            datainfo['Data_Name'] = 'ZDR'
        elif data_type_value == 34:
            datainfo['Data_Name'] = 'LDR'
        elif data_type_value == 35:
            datainfo['Data_Name'] = 'CC'
        elif data_type_value == 36:
            datainfo['Data_Name'] = 'DP'# ΦDP
        elif data_type_value == 37:
            datainfo['Data_Name'] = 'KDP'
        elif data_type_value == 38:
            datainfo['Data_Name'] = 'Re'
        elif data_type_value == 39:
            datainfo['Data_Name'] = 'VIL'
        elif data_type_value == 40:
            datainfo['Data_Name'] = 'HCL'
        elif data_type_value == 41:
            datainfo['Data_Name'] = 'SQI'
        elif data_type_value == 42:
            datainfo['Data_Name'] = 'CPA'
        elif data_type_value == 43:
            datainfo['Data_Name'] = 'CF'
        elif data_type_value == 44:
            datainfo['Data_Name'] = 'CP'
        elif data_type_value == 45:
            datainfo['Data_Name'] = 'BB'
        elif data_type_value == 46:
            datainfo['Data_Name'] = 'Cn2'
        elif 47 <= data_type_value <= 49:
            datainfo['Data_Name'] = 'Reserved'
        elif data_type_value == 50:
            datainfo['Data_Name'] = 'IWC'
        elif 51 <= data_type_value <= 64:
            datainfo['Data_Name'] = 'Reserved'
        yldbd['DataInfos'].append(datainfo)
        if(datainfo['Data_Type']>0):
            bsoffset_left = bsoffset_right
            bsoffset_right = bsoffset_left + datainfo['Bin_Bytes']*datainfo['Data_Length']
            data=np.frombuffer(bs[bsoffset_left:bsoffset_right],'i'+str(datainfo['Bin_Bytes']))
            data=np.where(data==0,nodata,(data-datainfo['Offset'])/datainfo['Scale'])
            yldbd['Data'].append(data)
        else:
            yldbd['Data'].append(None)
    varnames=list(map(lambda x: x['Data_Name'],yldbd.DataInfos))
    dvar = dict(zip(varnames, yldbd.Data))
    heights = np.arange(
        yldbd['CutConfigs'][0]['Start_Range'],
        yldbd['CutConfigs'][0]['Start_Range'] + yldbd['CutConfigs'][0]['Doppler_Resolution'] * yldbd.DataInfos[0][
            'Bin_Number'],
        yldbd['CutConfigs'][0]['Log_Resolution']
    )
    yldbd['Data'] = xr.Dataset(
        data_vars={key: (['time','height'], dvar[key][np.newaxis,:]) for key in dvar.keys()},
        coords={
            'time':[yldbd['TaskConfig']['Scan_Start_Time'].astype(datetime)],
            'height': heights
        },
        attrs={
            'time_count':1,
            'height_count': len(heights),
            'time_reference': 'UTC',
            'height_unit': 'meter',
        }
    )

    return yldbd

def BaseDatas_getDatas(self,fixData_Length='max',unobdata=unobdata)->xr.Dataset:
    '''
    获取云雷达基数据集中的数据
    Args:
        fixData_Length:不同时次数据对齐方式，max:最大长度,min:最小长度,或者指定长度
    Returns:
        xr.Dataset:数据
    '''
    varnames = list(map(lambda di: list(di.Data.data_vars), self.BaseDatas))
    varnames = list(set([item for sublist in varnames for item in sublist]))
    hcounts = list(map(lambda di: di.Data.height_count, self.BaseDatas))
    times = list(map(lambda di: di.Data.time.values[0], self.BaseDatas))
    if(fixData_Length=='max'):
        maxlendatas = max(hcounts)
    if(fixData_Length=='min'):
        maxlendatas = min(hcounts)
    if (isInt(fixData_Length)):
        maxlendatas = int(fixData_Length)
    heights = np.arange(
        self.BaseDatas[0]['CutConfigs'][0]['Start_Range'],
        self.BaseDatas[0]['CutConfigs'][0]['Start_Range'] + self.BaseDatas[0]['CutConfigs'][0][
            'Doppler_Resolution'] * maxlendatas,
        self.BaseDatas[0]['CutConfigs'][0]['Log_Resolution']
    )
    datavars = {}
    for key in varnames:
        dds = list(map(lambda x: np.squeeze(x.Data[key].values).tolist(), filter(lambda x: key in x.Data, self.BaseDatas)))
        ddas = [list(d)[:maxlendatas] if (len(d) > maxlendatas) else list(d) + [unobdata] * (maxlendatas - len(d)) for d
                in dds]
        datavars[key] = (['time', 'height'], np.array(ddas),)
    Datas = xr.Dataset(
        data_vars=datavars,
        coords={
            'time': times,
            'height': heights
        },
        attrs={
            'time_count':len(times),
            'height_count': maxlendatas,
            'time_reference': 'UTC',
            'height_unit': 'meter',
        }
    )
    self.Datas=Datas
BaseDatas.getDatas=BaseDatas_getDatas

def readBaseDatas(fps:list,use_multiprocess=False,multiproces_corenum=-1)->BaseDatas:
    '''    
    读取云雷达基数据文件列表
    Args:
        fps:文件路径列表
        use_multiprocess:是否使用多进程读取（速度较快但默认不使用）
        multiproces_corenum:多进程核心数（默认为-1，使用全部核心）
    Returns:
        list:云雷达对象列表    
    '''
    rbds=BaseDatas()
    if(use_multiprocess):
        rbds['BaseDatas']=Parallel(n_jobs=multiproces_corenum)(delayed(readSingleBaseData)(fp) for fp in fps)
    else:
        rbds['BaseDatas']=[readSingleBaseData(fp) for fp in fps]
    rbds.__datas__=rbds['BaseDatas']

    rbds.getDatas()
    return rbds

def BaseDatas_plot(self, data_name='Z1', figsize=(18,12), cmap=None, norm=None,show=True,savepath=None):
    '''    
    绘制云雷达数据    
    Args:
        plot_type:绘图类型,ref:反射率,velocity:速度,spectrumwith:谱宽,snr:信噪比
        figsize:图像尺寸
        cmap:颜色映射
        norm:归一化    
    '''
    if(data_name not in self.Datas.data_vars):
        raise ValueError('data_name not in Datas')

    if data_name=='Z1' or data_name=='Z2':
        if(cmap is None):
            cmap=ref_cmap
        if(norm is None):
            norm=ref_norm
    elif data_name=='V1' or data_name=='V2':
        unitstr='m/s'
        if(cmap is None):
            cmap=velocity_cmap
        if(norm is None):
            norm=velocity_norm
    elif data_name=='W1' or data_name=='W2':
        unitstr='m/s'
        if(cmap is None):
            cmap=spectrumwith_cmap
        if(norm is None):
            norm=spectrumwith_norm
    elif data_name=='SNR1' or data_name=='SNR2':
        unitstr='db'
        if(cmap is None):
            cmap=snr_cmap
        if(norm is None):
            norm=snr_norm

    else:
        unitstr='db'
        if(cmap is None):
            cmap=ref_cmap
        if(norm is None):
            norm=ref_norm

    fig,ax=plt.subplots(figsize=figsize)
    self.Datas[data_name].plot(
        ax=ax,
        x='time',
        cmap=cmap,
        norm=norm,
        cbar_kwargs=dict(
            orientation='horizontal',
            extend='max',
            extendrect=True,
            extendfrac='auto',
            pad=0.08,
            aspect=35
        )
    )
    ax.set_ylabel("Height (m)")
    ax.set_xlabel("Time (UTC)")
    if(show):
        plt.show()
    else:
        plt.close(fig)
    if(savepath is not None):
        fig.savefig(savepath,bbox_inches='tight')
BaseDatas.plot=BaseDatas_plot

def readStatuXMLfile(fp:str)->StatusData:
    '''
    读取毫米波云雷达雷达单个状态数据xml格式文件；
    支持 Z_RADA_I_IIiii_yyyyMMddhhmmss_R_YCCR_设备型号_STA_M.XML
    args:
        fp:单个状态数据xml格式文件
    return:
        StatusData对象
    '''
    try:
        with open(fp, 'r', encoding='utf8') as f:
            xml_data = f.read()
        xmld = ET.fromstring(xml_data)
        return StatusData.from_dict(parse_element(xmld))
    except Exception as ex:
        print(ex)
        return None

def readSingleStatuXMLfile(fp:str)->StatusData:
    '''
    读取毫米波云雷达雷达单个状态数据xml格式文件；
    支持 Z_RADA_I_IIiii_yyyyMMddhhmmss_R_YCCR_设备型号_STA_MM.XML
    args:
        fp:单个状态数据xml格式文件
    return:
        StatusData对象
    '''
    try:
        with open(fp, 'r', encoding='utf8') as f:
            xml_data = f.read()
        xmld = ET.fromstring(xml_data)
        return StatusData.from_dict(parse_element(xmld))
    except Exception as ex:
        print(ex)
        return None

def readSingleCalibrationXMLfile(fp:str)->CalibrationData:
    '''
    读取毫米波云雷达雷达单个标校数据xml格式文件；
    支持 Z_RADA_I_IIiii_yyyyMMddhhmmss_C_YCCR_设备型号_CAL.XML
    args:
        fp:单个标校数据xml格式文件
    return:
        CalibrationData对象
    '''
    try:
        with open(fp, 'r', encoding='utf8') as f:
            xml_data = f.read()
        xmld = ET.fromstring(xml_data)
        return CalibrationData.from_dict(parse_element(xmld))
    except Exception as ex:
        print(ex)
        return None



import numpy as np
import math

def isInt(istr:str)->bool:
    try:
        int(istr)
        return True
    except Exception as e:
        return False

def isFloat(istr:str)->bool:
    try:
        float(istr)
        return True
    except Exception as e:
        return False

def convert_bytes(size):
    power = 2**10
    n = 0
    units = {0: 'B', 1: 'KB', 2: 'MB', 3: 'GB', 4: 'TB'}
    while size > power:
        size /= power
        n += 1
    return f"{size:.2f} {units[n]}"

def dtryfloat(strf:str):
    try:
        return float(strf)
    except Exception as ex:
        return np.nan  
    
def uv2w(u, v):
    if(isFloat(u) and isFloat(v) and not np.isnan(u) and not np.isnan(v)):
        wdir = (180+math.atan2(u, v)/math.pi*180.0) % 360
        wspd = math.sqrt(u*u+v*v)
        return wdir, wspd
    else:
        return np.nan,np.nan

def w2uv(wdir, wspd):
    if(isFloat(wdir) and isFloat(wspd) and not np.isnan(wdir) and not np.isnan(wdir)):
        u = -wspd*math.sin(wdir/180*math.pi)
        v = -wspd*math.cos(wdir/180*math.pi)
        return u, v
    else:
        return np.nan,np.nan


def parse_element(element):
    if(len(element)>0):
        tags=list(map(lambda x:x.tag,element))
        # print(len(set(tags))==len(element))
        if(len(set(tags))==len(element)):
            parsed_data = {}
            for child in element:
                parsed_data[child.tag]=parse_element(child)
        else:
            parsed_data = []
            for child in element:
                parsed_data.append(parse_element(child))
    else:
        if(len(element.keys())>0):
            parsed_data= {attr: element.get(attr) for attr in element.keys()}
        else:
            parsed_data= element.text.strip()
    return parsed_data

vdtryfloat=np.vectorize(dtryfloat)
vw2uv=np.vectorize(w2uv)
vuv2w=np.vectorize(uv2w)


import os
import json
from datetime import datetime,timedelta
import dateutil.parser
import dateutil.rrule
import numpy as np
import pandas as pd
from . import Utils
import io
import math
import types
import functools
import matplotlib as mpl
import matplotlib.pyplot as plt

def readMWRFile(filename):
    ds=pd.read_csv(filename,encoding='gbk',skiprows=2)
    ds['DateTime']=pd.to_datetime(ds['DateTime'])
    ds['10']=ds['10'].astype(str).replace(
        '11',
        'TEM').replace(
        '12',
        'WVDen').replace(
        '13',
        'RHU').replace(
        '14',
        'WDen')
    ds=ds.rename(columns={'10':'dtype'})
    return ds



import numpy as np
import pandas as pd
import xarray as xr
import os
import os.path
import glob
import math
from datetime import datetime,timedelta
import dateutil.parser
import struct
import types
from joblib import Parallel, delayed
from . import base
originData=base.originData

L0Data=types.new_class('L0Data',(originData,))
L0Datas=types.new_class('L0Datas',(originData,))

def readSingleL0File(l0file:str)->L0Data:
    '''
    读取单个L0激光雷达文件
    Args:
        l0file: L0激光雷达文件路径

    Returns:
        L0Data: 激光雷达L0产品对象
    '''
    l0da=L0Data()
    dts=dateutil.parser.parse(l0file[-34:-20])    
    # dt=dts-timedelta(hours=8)
    l0da['Observe_Time']=dts
    dd=open(l0file,'rb')
    ds=dd.read()
    dd.close()
    pa=0
    channel_count=struct.unpack('H',ds[52:54])[0]
    if(channel_count!=8):
        pa=6
        channel_count=8
    l0da['Channel_Count']=channel_count
    channelmetas=[]
    for i in range(channel_count):
        channelmetas.append(struct.unpack('HHHHHHHH',ds[54+16*i+pa:70+16*i+pa]))
    params=pd.DataFrame(channelmetas,columns=['ID','WaveLength','type','Ratio','BHeight','PTR','CMethod','Count'])
    params['Ratio']=params['Ratio']/100
    params['BHeight']=params['BHeight']/10
    l0da['Channel_Metas']=params
    data=pd.DataFrame()
    for i in range(channel_count):
        data['c'+str(i+1)]=struct.unpack('f'*channelmetas[i][7],ds[channelmetas[i][5]:channelmetas[i][5]+4*channelmetas[i][7]])
    l0da['Data']=data
    return l0da

def readL0Files(fps:list,use_multiprocess=False,multiproces_corenum=-1)->L0Datas:
    '''    
    读取激光雷达文件列表    
    Args:
        fps:文件路径列表
        use_multiprocess:是否使用多进程读取（速度较快但默认不使用）
        multiproces_corenum:多进程核心数（默认为-1，使用全部核心）
    Returns:
        L0Datas:激光雷达对象列表    
    '''
    rbds=L0Datas()
    if(use_multiprocess):
        rbds['L0Datas']=Parallel(n_jobs=multiproces_corenum)(delayed(readSingleL0File)(fp) for fp in fps)
    else:
        rbds['L0Datas']=[readSingleL0File(fp) for fp in fps]
    rbds.__datas__=rbds['L0Datas']
    return rbds


#region 导出成厂家的DAT格式
def L0DataFileToManufactoryDat(l0file,outPath=None):
    l0da=readSingleL0File(l0file)
    data=l0da['Data']
    dts=l0da['Observe_Time']
    params=l0da['Channel_Metas']
    dt=dts-timedelta(hours=8)
    if(outPath==None):
        outPath='./'
    outf=outPath+dts.strftime('%Y/%Y%m%d/')+'S001-SLB001-Standard-'+str(int((dts-datetime(2023,1,1,0)).total_seconds()*2/3600+4275))\
        +'-'+dt.strftime('%y%m%d-%H%M%S')+'.dat'
    outp=open(outf,'w')
    outp.writelines(f'''File version 1.2
Local Time start:{dts.strftime('%H:%M:%S.000')}
Universal Time start:{dt.strftime('%Y-%m-%d %H:%M:%S.000')}
Various data: az=0.00;el=90.00;POWER=ALL_ON,,;Laser=Fire_ON,;LaserFreq=1000;StorageTimeUTC={dts.strftime('%H:%M:%S.000')};az_real=missing;el_real=missing;Servo=,,
Measure type = Standard
Scheduling = Sing=60;AzEnd=0;AzStart=0;ElEnd=90;ElStart=90;GlobScan=900;GlobRep=30;Chan=E355P,E355S,E532P,E532S,R386,R407,R607,E1064;RepRest=30;StepsAzim=1;StepsElev=1;PreTrig=10;T0_UT=102700.000;
Nota schedule =
Positioning = 0,0; 12; SLB001
Params: T=21; H=20; P=1002; R=15
Serie=12 Cycle=3
#######################################
     CHANNELS   PARAMS
E355P	E355S	R386	R407	E532P	E532S	R607	E1064	
''')
    outp.writelines('\t'.join(['%4d'%x for x in list(params['Count'].values)]))
    outp.writelines('''
100	100	100	100	100	100	100	100	
55000	55000	55000	55000	55000	55000	55000	55000	
DIG	DIG	DIG	DIG	DIG	DIG	DIG	DIG	
1000	1000	1000	1000	1000	1000	1000	1000	
-------------------------------------------
''')
    outp.writelines('''
'''.join(['\t'.join(['%d'%d for d in row]) for row in data.values]))
    outp.writelines('''
''')
    outp.close()
#endregion



def format_dict(data, indent_level=0):
    """
    手动格式化字典为字符串，支持处理包含子对象（如嵌套字典、列表等）的复杂结构，保证缩进正常。
    :param data: 要格式化的字典数据
    :param indent_level: 当前缩进级别，初始化为0，递归调用时会根据层级增加
    :return: 格式化后的字符串表示
    """
    indentstr='    '
    indent = indentstr* indent_level  # 根据缩进级别确定缩进空格
    formatted_str = "{\n"
    for key, value in data.items():
        if(key=='__datas__'):
            if(len(value)==0):
                continue
        formatted_str += f'{indent}{indentstr}"{key}": '
        if hasattr(value, '__dict__'):
            formatted_str += format_dict(dict(value.__dict__), indent_level + 1)+',\n'
        elif isinstance(value, list):
            # 如果值是列表，分别处理列表中的每个元素
            formatted_str += "[\n"
            for element in value:
                if hasattr(element, '__dict__'):
                    # 列表中的元素是字典，同样递归调用format_dict进行格式化，并增加缩进级别
                    formatted_str += f'{indent}{indentstr*2}'+format_dict(dict(element.__dict__), indent_level + 2) + ',\n'
                else:
                    # 列表中的非字典元素，直接转换为字符串并添加缩进
                    formatted_str += f'{indent}{indentstr*2}{repr(element)},\n'
            formatted_str = formatted_str.rstrip(",\n") + f'\n{indent}    ]\n'
        else:
            # 对于其他基本类型值，直接转换为字符串并添加缩进
            formatted_str += f'{repr(value)},\n'
    formatted_str = formatted_str.rstrip(",\n") + f'\n{indent}}}'
    return formatted_str

class originData():
    def __init__(self):
        self.__datas__=[]

    def __getitem__(self, key):
        if(isinstance(key, int)):
            return self.__datas__[key]
        else:
            return getattr(self, key)

    def __setitem__(self, key, value):
        if(isinstance(key, int)):
            self.__datas__[key]=value
        else:
            setattr(self, key, value)

    def append(self,value):
        self.__datas__.append(value)
    
    def __len__(self):
        return len(self.__datas__)

    def update(self, other):
        """
        更新当前originData实例，支持传入字典、originData实例、包含键值对元组的可迭代对象等进行更新。
        """
        if isinstance(other, dict):
            for key, value in other.items():
                self[key] = value
        elif isinstance(other, originData):
            for key in other:
                self[key] = other[key]
        elif hasattr(other, '__iter__'):  # 检查是否可迭代
            for element in other:
                if isinstance(element, tuple) and len(element) == 2:  # 检查是否为类似键值对的元组
                    key, value = element
                    self[key] = value

    def __str__(self):
        selfdict=dict(self.__dict__)
        if(self.__len__()==0):
            selfdict.pop('__datas__')
        return format_dict(selfdict)

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_dict(cls, input_dict):
        origin_obj = cls()
        for key, value in input_dict.items():
            origin_obj[key] = value
        return origin_obj
    
    def update(self,input_dict):
        for key, value in input_dict.items():
            self[key] = value



import os
import json
from datetime import datetime,timedelta
import dateutil.parser
import dateutil.rrule
import numpy as np
import pandas as pd
import xarray as xr
from . import Utils
import io
import math
import types
import matplotlib as mpl
from joblib import Parallel,delayed
import xml.etree.ElementTree as ET
from . import base,Utils
originData=base.originData
parse_element=Utils.parse_element

#region 绘图参数
velocity_colors=[
    '#951262',
    '#ED2226',
    '#EB3E22',
    '#EF6A26',
    '#F58324',
    '#F3A122',
    '#FDB52A',
    '#FDD31E',
    '#F9EF1A',
    '#F7FD1C',
    '#D9E9F1',
    '#C7D7F1',
    '#B1C7EB',
    '#99B5E7',
    '#83A5DD',
    '#6E93D3',
    '#6089CB',
    '#4C7CC3',
    '#4870BB',
    '#406AB9',
    '#2A42AB',
    '#242681',
    '#242074',
    '#22206A',
    '#1C622C',
    '#067E42',
    '#0C8B42',
    '#30BD4C',
    '#4EC14A',
    '#58C546',
    '#87742E'
]
velocity_levels=[-1,-0.9,-0.8,-0.7,-0.6,-0.5,-0.4,-0.3,-0.1,0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0,2,2.5,3,4,5,6,7,8,9,10]
velocity_cmap = (mpl.colors.ListedColormap(velocity_colors).with_extremes(over=velocity_colors[-1], under=velocity_colors[0]))
velocity_norm = mpl.colors.BoundaryNorm(velocity_levels, velocity_cmap.N)
#endregion

#region 风廓线雷达数据文件元数据
# 定义变量名字典
line0_variables = {
    'WNDRAD':'风廓线雷达数据文件标识',
    'File_version':'文件版本号'
}
line1_variables = {
    'Station_Id':'站号',
    'Longitude':'经度',
    'Latitude':'纬度',
    'Altitude':'海拔高度',
    'Machine_Type':'风廓线仪型号',
}
line2_variables = {
    'Antenna_Gain': '天线增益',
    'Feeder_Loss': '馈线损耗',
    'East_Beam_Angle': '东波束与铅垂线的夹角',
    'West_Beam_Angle': '西波束与铅垂线的夹角',
    'South_Beam_Angle': '南波束与铅垂线的夹角',
    'North_Beam_Angle': '北波束与铅垂线的夹角',
    'Center_Row_Beam_Angle': '中（行）波束与铅垂线的夹角（度）',
    'Center_Column_Beam_Angle': '中（列）波束与铅垂线的夹角（度）',
    'Number_Of_Beams': '波束数',
    'Sampling_Frequency': '采样频率',
    'Transmit_Wavelength': '发射波长',
    'Pulse_Repetition_Frequency': '脉冲重复频率',
    'Pulse_Width': '脉冲宽度',
    'Horizontal_Beam_Width': '水平波束宽度',
    'Vertical_Beam_Width': '垂直波束宽度',
    'Peak_Transmit_Power': '发射峰值功率',
    'Average_Transmit_Power': '发射平均功率',
    'Start_Sampling_Height': '起始采样高度',
    'End_Sampling_Height': '终止采样高度'
}
line3_variables = {
    'Time_Source': '时间来源',
    'Observation_Start_Time': '观测开始时间',
    'Observation_End_Time': '观测结束时间',
    'Calibration_Status': '标校状态',
    'Incoherent_Accumulation': '非相干积累',
    'Coherent_Accumulation': '相干积累',
    'FFT_Points': 'Fft点数',
    'Spectral_Averages': '谱平均数',
    'Beam_Order_Flag': '波束顺序标志',
    'East_Beam_Azimuth_Correction': '东波束方位角修正值',
    'West_Beam_Azimuth_Correction': '西波束方位角修正值',
    'South_Beam_Azimuth_Correction': '南波束方位角修正值',
    'North_Beam_Azimuth_Correction': '北波束方位角修正值'
}
L2data_variables={
    'Sampling_Height': '采样高度',
    'Velocity_Spectrum_Width': '速度谱宽',
    'Signal_To_Noise_Ratio': '信噪比',
    'Radial_Velocity': '径向速度'
}
L3meta_variables={
    "Station_Id": "区站号",
    "Longitude": "经度",
    "Latitude": "纬度",
    "Altitude": "观测场拔海高度",
    "Machine_Type": "风廓线仪型号",
    "Observation_Time": "观测时间"
}
L3data_variables = {
    "Sampling_Height": "采样高度",
    "Wind_Direction": "水平风向",
    "Wind_Speed": "水平风速",
    "Vertical_Wind_Speed": "垂直风速",
    "Horizontal_Confidence": "水平方向可信度",
    "Vertical_Confidence": "垂直方向可信度",
    "Cn2": "Cn2"
} 
L1Struct_FileTag = np.dtype([
    ('FileID', 'S8'),   # 字符类型数组，用于存储文件标识，字节数为8，这里固定值为WNDFFT
    ('VersionNo', 'f4'),  # 单精度浮点数类型，用于表示数据格式版本号，字节数为4，格式为两位整数两位小数，示例值如01.20
    ('FileHeaderLength', 'i4') # 32位整数类型，用于表示文件头的长度，字节数为4，是4位整数
])
L1Struct_SiteInfo = np.dtype([
    ('Country', 'S16'),
    ('Province', 'S16'),
    ('StationNumber', 'S16'),
    ('Station', 'S16'),
    ('RadarType', 'S16'),
    ('Longitude', 'S16'),
    ('Latitude', 'S16'),
    ('Altitude', 'S16'),
    ('Temp', 'S40')
])
L1Struct_PerformanceInfo = np.dtype([
    ('Ae', 'i4'),
    ('AgcWast', 'f4'),
    ('AngleE', 'f4'),
    ('AngleW', 'f4'),
    ('AngleS', 'f4'),
    ('AngleN', 'f4'),
    ('AngleR', 'f4'),
    ('AngleL', 'f4'),
    ('ScanBeamN', 'u4'),
    ('SampleP', 'u4'),
    ('WaveLength', 'u4'),
    ('Prp', 'f4'),
    ('PusleW', 'f4'),
    ('HBeamW', 'u2'),
    ('VBeamW', 'u2'),
    ('TranPp', 'f4'),
    ('TranAp', 'f4'),
    ('StartSamplBin', 'u4'),
    ('EndSamplBin', 'u4'),
    ('BinLength', 'i2'),
    ('BinNum', 'i2'),
    ('Temp', 'S40')
])
L1Struct_ObservationInfo = np.dtype([
    ('SYear', 'u2'),
    ('SMonth', 'u1'),
    ('SDay', 'u1'),
    ('SHour', 'u1'),
    ('SMinute', 'u1'),
    ('SSecond', 'u1'),
    ('TimeP', 'u1'),
    ('SMillisecond', 'u4'),
    ('Calibration', 'u2'),
    ('BeamfxChange', 'i2'),
    ('EYear', 'u2'),
    ('EMonth', 'u1'),
    ('EDay', 'u1'),
    ('EHour', 'u1'),
    ('EMinute', 'u1'),
    ('ESecond', 'u1'),
    ('NNtr', 'i1'),
    ('Ntr', 'i2'),
    ('SpAver', 'u2'),
    ('Fft', 'u2'),
    ('unknow', 'u2'),
    ('BeamDir', 'S12'),
    ('AzimuthE', 'f4'),
    ('AzimuthW', 'f4'),
    ('AzimuthS', 'f4'),
    ('AzimuthN', 'f4'),
    ('Temp', 'S40')
])
#endregion


L1Data=types.new_class('L1Data',(originData,))
L1ObData=types.new_class('L1ObData',(originData,))
L2Data=types.new_class('L2Data',(originData,))
L2LevelData=types.new_class('L2LevelData',(originData,))
L3Data=types.new_class('L3Data',(originData,))
L3Datas=types.new_class('L3Datas',(originData,))
CalibrationData=types.new_class('CalibrationDatas',(originData,))
StatusData=types.new_class('StatusData',(originData,))


def readSingleL1file(fp:str)->L1Data:
    '''
    读取风廓线雷达单个L1功率谱文件，文件命名格式为：Z_RADA_I_IIiii_yyyyMMddhhmmss_O_WPRD_雷达型号_FFT.BIN
    args:
        fp:L1功率谱文件路径
    return:
        L1Data对象
    '''
    try:
        with open(fp, 'rb') as f:
            bs=f.read()
        fkxL1obj=L1Data()
        bsoffset = 0
        data = np.frombuffer(bs[bsoffset:bsoffset + L1Struct_FileTag.itemsize], L1Struct_FileTag)
        bsoffset = bsoffset + L1Struct_FileTag.itemsize
        fkxL1obj['FileTag'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
        fkxL1obj.FileTag['FileID'] = fkxL1obj.FileTag['FileID'].split(b'\x00')[0].decode('gbk')
        data = np.frombuffer(bs[bsoffset:bsoffset + L1Struct_SiteInfo.itemsize], L1Struct_SiteInfo)
        bsoffset = bsoffset + L1Struct_SiteInfo.itemsize
        fkxL1obj['SiteInfo'] = originData.from_dict(
            {name: data[name][0].split(b'\x00')[0].decode('gbk') for name in data.dtype.names}
        )
        fkxL1obj['Datas'] = []
        while (bsoffset < len(bs)):
            L1ObDatai = L1ObData()
            data = np.frombuffer(bs[bsoffset:bsoffset + L1Struct_PerformanceInfo.itemsize], L1Struct_PerformanceInfo)
            bsoffset = bsoffset + L1Struct_PerformanceInfo.itemsize
            L1ObDatai['PerformanceInfo'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
            L1ObDatai.PerformanceInfo['Temp'] = L1ObDatai.PerformanceInfo['Temp'].split(b'\x00')[0].decode('gbk')
            data = np.frombuffer(bs[bsoffset:bsoffset + L1Struct_ObservationInfo.itemsize], L1Struct_ObservationInfo)
            bsoffset = bsoffset + L1Struct_ObservationInfo.itemsize
            L1ObDatai['ObservationInfo'] = originData.from_dict({name: data[name][0] for name in data.dtype.names})
            L1ObDatai.ObservationInfo.BeamDir = L1ObDatai.ObservationInfo['BeamDir'].decode('gbk')
            L1ObDatai.ObservationInfo.Temp = L1ObDatai.ObservationInfo['Temp'].decode('gbk')
            L1ObDatai.Speed_per_FFT_point = L1ObDatai.PerformanceInfo.Prp / 2 / L1ObDatai.ObservationInfo.Fft * L1ObDatai.PerformanceInfo.WaveLength * 1e-5
            data_len = L1ObDatai.ObservationInfo.Fft * L1ObDatai.PerformanceInfo.BinNum * L1ObDatai.PerformanceInfo.ScanBeamN
            data = np.frombuffer(
                bs[bsoffset:bsoffset + 4 * data_len],
                'f4',
                count=data_len
            ).reshape(L1ObDatai.PerformanceInfo.ScanBeamN, L1ObDatai.PerformanceInfo.BinNum,
                      L1ObDatai.ObservationInfo.Fft)
            bsoffset = bsoffset + 4 * data_len
            da = xr.DataArray(
                data,
                dims=['Beam', 'height', 'Fft'],
                coords={
                    'Beam': list(L1ObDatai.ObservationInfo.BeamDir),
                    'height': np.arange(
                        L1ObDatai.PerformanceInfo.BinNum) * L1ObDatai.PerformanceInfo.BinLength + L1ObDatai.PerformanceInfo.StartSamplBin,
                    'Fft': np.arange(L1ObDatai.ObservationInfo.Fft)
                },
                attrs={
                    'WaveLength': L1ObDatai.PerformanceInfo.WaveLength,
                    'Prp': L1ObDatai.PerformanceInfo.Prp,
                    'ObStarTime': datetime(
                        L1ObDatai.ObservationInfo.SYear,
                        L1ObDatai.ObservationInfo.SMonth,
                        L1ObDatai.ObservationInfo.SDay,
                        L1ObDatai.ObservationInfo.SHour,
                        L1ObDatai.ObservationInfo.SMinute,
                        L1ObDatai.ObservationInfo.SSecond,
                    ),
                    'ObEndTime': datetime(
                        L1ObDatai.ObservationInfo.EYear,
                        L1ObDatai.ObservationInfo.EMonth,
                        L1ObDatai.ObservationInfo.EDay,
                        L1ObDatai.ObservationInfo.EHour,
                        L1ObDatai.ObservationInfo.EMinute,
                        L1ObDatai.ObservationInfo.ESecond,
                    ),
                    'Speed_per_FFT_point': L1ObDatai.Speed_per_FFT_point,
                    'WaveLength_unit': '1e-5 meter',
                    'Prp_unit': 'Hz',
                    'Speed_per_FFT_point_unit': 'm/s',
                    'height_unit': 'meter',
                }
            )
            L1ObDatai['DspToDpDat'] = da
            fkxL1obj['Datas'].append(L1ObDatai)
        return fkxL1obj
    except Exception as ex:
        print(ex)
        return None

def readSingleL2file(fp:str)->L2Data:
    '''
    读取风廓线雷达单个L2径向数据文件，文件命名格式为：Z_RADA_I_IIiii_yyyyMMddhhmmss_O_WPRD_雷达型号_RAD.TXT
    args:
        fp:L2产品径向数据文件路径
    return:
        L2Data对象
    '''
    try:
        with open(fp, 'r') as f:
            lines = [line.strip() for line in f]        
        fkxL2obj=L2Data()

        lineds=lines[0].split(' ')    
        keys=list(line0_variables.keys())    
        for i, key in enumerate(keys):
            fkxL2obj[key] = lineds[i]

        lineds=lines[1].split(' ')
        keys=list(line1_variables.keys())
        for i, key in enumerate(keys):
            fkxL2obj[key] = lineds[i] if i == 0 or i == 4 else Utils.dtryfloat(lineds[i])

        fkxL2obj['levels']=[]

        tsplitlineis = [1] + [i for i in range(5, len(lines)) if lines[i] == 'NNNN']

        columnnames=list(L2data_variables.keys())
        leveli=None
        beamj=0
        Beam_Count=1
        dfleveli=[]
        for i in range(len(tsplitlineis)-1):
            lineStart=tsplitlineis[i]
            lineEnd=tsplitlineis[i+1]  
            if(i%Beam_Count==0):
                leveli=L2LevelData()
                dfleveli=[]
                beamj=0
                lineStart=lineStart+1
                lineds=lines[lineStart].split(' ')
                keys=list(line2_variables.keys())
                for j in range(len(keys)):
                    key=keys[j]
                    leveli[key]=Utils.dtryfloat(lineds[j])

                lineStart=lineStart+1
                lineds=lines[lineStart].split(' ')
                keys=list(line3_variables.keys())
                for j, key in enumerate(keys):
                    if(key=='Beam_Order_Flag'):
                        leveli[key]=lineds[j].replace('/','')
                        leveli['Beam_Order_Flags']=list(map(lambda bi:leveli['Beam_Order_Flag'][bi],range(len(leveli['Beam_Order_Flag']))))
                        leveli['Beam_Count']=len(leveli['Beam_Order_Flag'])
                    else:
                        if(key=='Observation_Start_Time' or key=='Observation_End_Time'):
                            leveli[key]=dateutil.parser.parse(lineds[j])
                        else:           
                            leveli[key]=Utils.dtryfloat(lineds[j])
                Beam_Count=leveli['Beam_Count']

            dfleveli.append(lines[lineStart+2:lineEnd])
            beamj=beamj+1
            
            if(beamj==Beam_Count):            
                # 顺序逐层添加（不包含末尾层）
                dlevelcolumnnames = [
                    f"{beam_order_flag}_{key}" for beam_order_flag in leveli['Beam_Order_Flags'] for key in columnnames  
                ] 
                sscolumnnames=[dlevelcolumnnames[i] for i in range(0,len(dlevelcolumnnames),len(columnnames))]
                dfleveli=np.array(dfleveli).T
                dfleveli=pd.read_csv(io.StringIO('\n'.join([' '.join(x) for x in list(dfleveli)])),sep=' ',header=None,names=dlevelcolumnnames)
                dfleveli=dfleveli.rename(columns={sscolumnnames[0]:sscolumnnames[0][2:]}).drop(sscolumnnames[1:],axis=1)
                dfleveli=dfleveli.apply(lambda x:pd.to_numeric(x,errors='coerce'))   
                leveli['Data']=dfleveli
                fkxL2obj['levels'].append(leveli)

        
        fkxL2obj['Beam_Order_Flags']=fkxL2obj['levels'][0]['Beam_Order_Flags']
        fkxL2obj['Beam_Count']=fkxL2obj['levels'][0]['Beam_Count']
        levels_count=len(fkxL2obj['levels'])
        fkxL2obj['levels_count']=levels_count
        if(levels_count==1):
            dlevels=['low']
        else:
            if(levels_count==2):
                dlevels=['low','high']
            else:
                dlevels=['low','middle','high']
        fkxL2obj['dlevels']=dlevels

        return fkxL2obj
    except Exception as ex:
        print(ex)
        # raise ex
        return None

def CalcL2toL3(fkxL2obj,qcw=3,interp=False,rollmean=True,rollmeancout=5):
    '''
    基于L2数据数学计算L3数据
    args:
        fkxL2obj:L2Data对象
        qcw:质控阈值
        interp:是否插值
        rollmean:是否滚动平均
        rollmeancout:滚动平均窗口大小
    return:
        L3Data对象
    '''

    Calc_L2_data=[]
    for j,levelj in enumerate(fkxL2obj.levels):
        dlevelj=levelj.Data
        qcindexs=dlevelj[dlevelj['E_Radial_Velocity']+dlevelj['W_Radial_Velocity']>qcw].index
        dlevelj.loc[qcindexs,'E_Radial_Velocity']=np.nan
        dlevelj.loc[qcindexs,'W_Radial_Velocity']=np.nan
        qcindexs=dlevelj[dlevelj['N_Radial_Velocity']+dlevelj['S_Radial_Velocity']>qcw].index
        dlevelj.loc[qcindexs,'N_Radial_Velocity']=np.nan
        dlevelj.loc[qcindexs,'S_Radial_Velocity']=np.nan

        h=dlevelj['Sampling_Height'].values
        Vre=dlevelj['E_Radial_Velocity'].values
        # Wre=dlevelj['E_Signal_To_Noise_Ratio'].values/np.exp(dlevelj['E_Velocity_Spectrum_Width'].values*20)
        Wre=1/np.exp(dlevelj['E_Velocity_Spectrum_Width'].values)
        Vrn=dlevelj['N_Radial_Velocity'].values
        # Wrn=dlevelj['N_Signal_To_Noise_Ratio'].values/np.exp(dlevelj['N_Velocity_Spectrum_Width'].values*20)
        Wrn=1/np.exp(dlevelj['N_Velocity_Spectrum_Width'].values)
        Vw=dlevelj['R_Radial_Velocity'].values
        # Ww=dlevelj['R_Signal_To_Noise_Ratio'].values/np.exp(dlevelj['R_Velocity_Spectrum_Width'].values*20)
        Ww=1/np.exp(dlevelj['R_Velocity_Spectrum_Width'].values)
        Vrw=dlevelj['W_Radial_Velocity'].values
        # Wrw=dlevelj['W_Signal_To_Noise_Ratio'].values/np.exp(dlevelj['W_Velocity_Spectrum_Width'].values*20)
        Wrw=1/np.exp(dlevelj['W_Velocity_Spectrum_Width'].values)
        Vrs=dlevelj['S_Radial_Velocity'].values
        # Wrs=dlevelj['S_Signal_To_Noise_Ratio'].values/np.exp(dlevelj['S_Velocity_Spectrum_Width'].values*20)
        Wrs=1/np.exp(dlevelj['S_Velocity_Spectrum_Width'].values)
        Vs=np.array([Vre,Vrw,Vrs,Vrn,Vw])
        Ws=np.array([Wre,Wrw,Wrs,Wrn,Ww])

        Vsw=np.array([
            np.where(np.isnan(Vrw),np.nan,Vre),np.where(np.isnan(Vre),np.nan,Vrw),
            np.where(np.isnan(Vrn),np.nan,Vrs),np.where(np.isnan(Vrs),np.nan,Vrn),
            Vw
        ])
        w=np.sum(((Vsw*Ws)/np.nansum(Ws,axis=0)),axis=0)
        # w=np.nanmean(Vsw,axis=0)

        a_e=levelj['East_Beam_Angle']+levelj['East_Beam_Azimuth_Correction']
        o_e=90-a_e
        po_e=math.radians(o_e)
        u_e=(w*math.sin(po_e)-Vre)/math.cos(po_e)
        a_w=levelj['West_Beam_Angle']+levelj['West_Beam_Azimuth_Correction']
        o_w=90-a_w
        po_w=math.radians(o_w)
        u_w=(Vrw-w*math.sin(po_w))/math.cos(po_w)    
        u=np.sum((np.array([u_e,u_w])*Ws[:2])/np.nansum(Ws[:2],axis=0),axis=0) 

        a_s=levelj['South_Beam_Angle']+levelj['South_Beam_Azimuth_Correction']
        o_s=90-a_s
        po_s=math.radians(o_s)
        v_s=(Vrs-w*math.sin(po_s))/math.cos(po_s)
        a_n=levelj['North_Beam_Angle']+levelj['North_Beam_Azimuth_Correction']
        o_n=90-a_n
        po_n=math.radians(o_n)
        v_n=(w*math.sin(po_n)-Vrn)/math.cos(po_n)   
        v=np.sum((np.array([v_s,v_n])*Ws[2:4])/np.nansum(Ws[2:4],axis=0),axis=0)    

        calcdfi=pd.DataFrame({
            'Sampling_Height':h,
            'Vertical_Wind_Speed':w,
            'U_Wind_Speed':u,
            # 'U_Wind_Speede':u_e,
            # 'U_Wind_Speedw':u_w,
            'V_Wind_Speed':v,
            # 'V_Wind_Speedn':v_n,
            # 'V_Wind_Speeds':v_s
        })
        Calc_L2_data.append(calcdfi)  
    Calc_L3Data=pd.concat(Calc_L2_data).reset_index(drop=True).groupby('Sampling_Height').mean().reset_index()
    if(interp):
        Calc_L3Data=Calc_L3Data.set_index('Sampling_Height').interpolate(method='linear').reset_index()
    if(rollmean):        
        Calc_L3Data=Calc_L3Data.set_index('Sampling_Height').sort_index().rolling(window=rollmeancout, min_periods=1).mean().reset_index()
    wdir,wspd=Utils.vuv2w(Calc_L3Data['U_Wind_Speed'].values,Calc_L3Data['V_Wind_Speed'].values)
    Calc_L3Data['Wind_Direction']=wdir
    Calc_L3Data['Wind_Speed']=wspd
    Calc_L3Data=Calc_L3Data[['Sampling_Height', 'Wind_Direction', 'Wind_Speed', 'Vertical_Wind_Speed', 'U_Wind_Speed','V_Wind_Speed']]
    Calc_L3Data.dropna(inplace=True)   
    
    fkxL3obj=L3Data()
    keys=list(line0_variables.keys())    
    for key in keys:
        fkxL3obj[key]=fkxL2obj[key]
    keys=list(L3meta_variables.keys())    
    for key in keys:
        if(key=='Observation_Time'):
            fkxL3obj[key]=fkxL2obj.levels[0]['Observation_End_Time']     
        else:
            fkxL3obj[key]=fkxL2obj[key]      
    fkxL3obj['Data']=Calc_L3Data
    
    return fkxL3obj

def readSingleL3file(fp:str)->L3Data:
    '''
    读取风廓线雷达单个L3产品数据文件；
    支持
    ROBS：Z_RADA_I_IIiii_yyyyMMddhhmmss_P_WPRD_雷达型号_ROBS.TXT
    HOBS：Z_RADA_I_IIiii_yyyyMMddhhmmss_P_WPRD_雷达型号_HOBS.TXT
    OOBS：Z_RADA_I_IIiii_yyyyMMddhhmmss_P_WPRD_雷达型号_OOBS.TXT
    args:
        fp:L3产品文件路径
    return:
        L3Data对象
    '''
    try:
        with open(fp, 'r') as f:
            lines = [line.strip() for line in f]      

        fkxL3obj=L3Data()    
        lineds=lines[0].split(' ')    
        keys=list(line0_variables.keys())    
        for i, key in enumerate(keys):
            fkxL3obj[key] = lineds[i]

        lineds=lines[1].split(' ')
        keys=list(L3meta_variables.keys())
        for i, key in enumerate(keys):
            if(key=='Station_Id' or key=='Machine_Type'):            
                fkxL3obj[key]=lineds[i]
            else:
                if(key=='Observation_Time'):
                    fkxL3obj[key]=dateutil.parser.parse(lineds[i])
                else:
                    fkxL3obj[key]=float(lineds[i])

        da=pd.read_fwf(
            io.StringIO('\n'.join(lines[3:-1])), 
            widths=[5, 6, 6, 7, 4, 4, 9],
            encoding='gbk',
            names=list(L3data_variables.keys()),
            dtype=str
        )
        # 之前定义各列使用 str 类型存储，后续需要把 //// 表示的缺失值用 numpy.nan 进行替换
        da=da.apply(lambda x:pd.to_numeric(x,errors='coerce'))  
        uvs = Utils.vw2uv(da['Wind_Direction'].values, da['Wind_Speed'].values)
        da['U_Wind_Speed']=uvs[0]
        da['V_Wind_Speed']=uvs[1]
        fkxL3obj['Data']=da
        return fkxL3obj
    except Exception as ex:
        print(ex)
        # raise ex
        return None

def readL3files(fps:list,use_multiprocess=False,multiproces_corenum=-1)->L3Datas:
    """
    读取多个L3产品文件，并返回L3Datas对象。

    参数:
        fps (list): L3产品文件路径列表。
        use_multiprocess (bool): 是否使用多进程读取文件，默认为False。
        multiproces_corenum (int): 多进程核心数量，默认为-1表示使用所有可用核心。

    返回:
        L3Datas: 包含所有读取的L3Data对象的L3Datas实例。
    """
    rbds=L3Datas()
    if(use_multiprocess):
        rbds['L3Datas']=Parallel(n_jobs=multiproces_corenum)(delayed(readSingleL3file)(fp) for fp in fps)
    else:
        rbds['L3Datas']=[readSingleL3file(fp) for fp in fps]
    for i,ld in enumerate(rbds['L3Datas']):
        rbds.append(ld)    
    return rbds

def readSingleCalibrationXMLfile(fp:str)->CalibrationData:
    '''
    读取风廓线雷达单个标校数据xml格式文件；
    支持 Z_RADA_I _IIiii_yyyyMMddhhmmss_C_WPRD_雷达型号_CAL.XML
    args:
        fp:单个标校数据xml格式文件
    return:
        CalibrationData对象
    '''
    try:
        with open(fp, 'r', encoding='utf8') as f:
            xml_data = f.read()
        xmld = ET.fromstring(xml_data)
        return CalibrationData.from_dict(parse_element(xmld))
    except Exception as ex:
        print(ex)
        return None

def readSingleStatuXMLfile(fp:str)->StatusData:
    '''
    读取风廓线雷达单个状态数据xml格式文件；
    支持 Z_RADA_I_IIiii_yyyyMMddhhmmss_R_WPRD_雷达型号_STA.XML
    args:
        fp:单个状态数据xml格式文件
    return:
        StatusData对象
    '''
    try:
        with open(fp, 'r', encoding='utf8') as f:
            xml_data = f.read()
        xmld = ET.fromstring(xml_data)
        return StatusData.from_dict(parse_element(xmld))
    except Exception as ex:
        print(ex)
        return None
